/* This file is autogenerated from the database. */
/* See src/desc/gen_proto */
/* Do not edit*/
entree functions_default[]={
{"TeXstyle",0,(void*)sd_TeXstyle,22,"",""},
{"breakloop",0,(void*)sd_breakloop,22,"",""},
{"colors",0,(void*)sd_colors,22,"",""},
{"compatible",0,(void*)sd_compatible,22,"",""},
{"datadir",0,(void*)sd_datadir,22,"",""},
{"debug",0,(void*)sd_debug,22,"",""},
{"debugfiles",0,(void*)sd_debugfiles,22,"",""},
{"debugmem",0,(void*)sd_debugmem,22,"",""},
{"echo",0,(void*)sd_echo,22,"",""},
{"factor_add_primes",0,(void*)sd_factor_add_primes,22,"",""},
{"factor_proven",0,(void*)sd_factor_proven,22,"",""},
{"factorlimit",0,(void*)sd_factorlimit,22,"",""},
{"format",0,(void*)sd_format,22,"",""},
{"graphcolormap",0,(void*)sd_graphcolormap,22,"",""},
{"graphcolors",0,(void*)sd_graphcolors,22,"",""},
{"help",0,(void*)sd_help,22,"",""},
{"histfile",0,(void*)sd_histfile,22,"",""},
{"histsize",0,(void*)sd_histsize,22,"",""},
{"lines",0,(void*)sd_lines,22,"",""},
{"linewrap",0,(void*)sd_linewrap,22,"",""},
{"log",0,(void*)sd_log,22,"",""},
{"logfile",0,(void*)sd_logfile,22,"",""},
{"nbthreads",0,(void*)sd_nbthreads,22,"",""},
{"new_galois_format",0,(void*)sd_new_galois_format,22,"",""},
{"output",0,(void*)sd_output,22,"",""},
{"parisize",0,(void*)sd_parisize,22,"",""},
{"parisizemax",0,(void*)sd_parisizemax,22,"",""},
{"path",0,(void*)sd_path,22,"",""},
{"plothsizes",0,(void*)sd_plothsizes,22,"",""},
{"prettyprinter",0,(void*)sd_prettyprinter,22,"",""},
{"primelimit",0,(void*)sd_primelimit,22,"",""},
{"prompt",0,(void*)sd_prompt,22,"",""},
{"prompt_cont",0,(void*)sd_prompt_cont,22,"",""},
{"psfile",0,(void*)sd_psfile,22,"",""},
{"readline",0,(void*)sd_readline,22,"",""},
{"realbitprecision",0,(void*)sd_realbitprecision,22,"",""},
{"realprecision",0,(void*)sd_realprecision,22,"",""},
{"recover",0,(void*)sd_recover,22,"",""},
{"secure",0,(void*)sd_secure,22,"",""},
{"seriesprecision",0,(void*)sd_seriesprecision,22,"",""},
{"simplify",0,(void*)sd_simplify,22,"",""},
{"sopath",0,(void*)sd_sopath,22,"",""},
{"strictargs",0,(void*)sd_strictargs,22,"",""},
{"strictmatch",0,(void*)sd_strictmatch,22,"",""},
{"threadsize",0,(void*)sd_threadsize,22,"",""},
{"threadsizemax",0,(void*)sd_threadsizemax,22,"",""},
{"timer",0,(void*)sd_timer,22,"",""},
{NULL,0,NULL,0,NULL,NULL} /* sentinel */
};
