#ifdef LONG_IS_64BIT
#define __AGM_ATAN_LIMIT                 3584
#define __DIVRR_GMP_LIMIT                -1
#define __EXPNEWTON_LIMIT                4224
#define __EXTGCD_HALFGCD_LIMIT           70
#define __F2x_MUL_KARATSUBA_LIMIT        11
#define __F2x_MUL_MULII_LIMIT            662
#define __F2xqX_BARRETT_LIMIT            56
#define __F2xqX_DIVREM_BARRETT_LIMIT     127
#define __F2xqX_EXTGCD_LIMIT             10
#define __F2xqX_GCD_LIMIT                468
#define __F2xqX_HALFGCD_LIMIT            64
#define __F2xqX_INVBARRETT_LIMIT         50
#define __F2xqX_REM_BARRETT_LIMIT        127
#define __Flx_BARRETT2_LIMIT             63
#define __Flx_BARRETT_LIMIT              131
#define __Flx_DIVREM2_BARRETT_LIMIT      28
#define __Flx_DIVREM_BARRETT_LIMIT       407
#define __Flx_EXTGCD2_LIMIT              31
#define __Flx_EXTGCD_LIMIT               17
#define __Flx_GCD2_LIMIT                 2266
#define __Flx_GCD_LIMIT                  621
#define __Flx_HALFGCD2_LIMIT             32
#define __Flx_HALFGCD_LIMIT              265
#define __Flx_INVBARRETT2_LIMIT          28
#define __Flx_INVBARRETT_LIMIT           316
#define __Flx_MUL2_KARATSUBA_LIMIT       10
#define __Flx_MUL2_MULII_LIMIT           14
#define __Flx_MUL_KARATSUBA_LIMIT        38
#define __Flx_MUL_MULII_LIMIT            66
#define __Flx_REM2_BARRETT_LIMIT         241
#define __Flx_REM_BARRETT_LIMIT          407
#define __Flx_SQR2_KARATSUBA_LIMIT       15
#define __Flx_SQR2_SQRI_LIMIT            28
#define __Flx_SQR_KARATSUBA_LIMIT        94
#define __Flx_SQR_SQRI_LIMIT             77
#define __FlxqX_BARRETT_LIMIT            17
#define __FlxqX_DIVREM_BARRETT_LIMIT     46
#define __FlxqX_EXTGCD_LIMIT             10
#define __FlxqX_GCD_LIMIT                407
#define __FlxqX_HALFGCD_LIMIT            57
#define __FlxqX_INVBARRETT_LIMIT         22
#define __FlxqX_REM_BARRETT_LIMIT        48
#define __FpXQX_BARRETT_LIMIT            12
#define __FpXQX_DIVREM_BARRETT_LIMIT     30
#define __FpXQX_EXTGCD_LIMIT             10
#define __FpXQX_GCD_LIMIT                152
#define __FpXQX_HALFGCD_LIMIT            43
#define __FpXQX_INVBARRETT_LIMIT         40
#define __FpXQX_REM_BARRETT_LIMIT        30
#define __FpX_BARRETT_LIMIT              85
#define __FpX_DIVREM_BARRETT_LIMIT       292
#define __FpX_EXTGCD_LIMIT               21
#define __FpX_GCD_LIMIT                  282
#define __FpX_HALFGCD_LIMIT              56
#define __FpX_INVBARRETT_LIMIT           254
#define __FpX_REM_BARRETT_LIMIT          306
#define __Fp_POW_BARRETT_LIMIT           101
#define __Fp_POW_REDC_LIMIT              99
#define __GCD_HALFGCD_LIMIT              9
#define __HALFGCD_LIMIT                  85
#define __INVMOD_GMP_LIMIT               -1
#define __INVNEWTON_LIMIT                35200
#define __LOGAGMCX_LIMIT                 832
#define __LOGAGM_LIMIT                   1024
#define __MULII_FFT_LIMIT                1441
#define __MULII_KARATSUBA_LIMIT          23
#define __MULRR_MULII_LIMIT              13440
#define __RgX_MUL_LIMIT                  9
#define __RgX_SQR_LIMIT                  35
#define __SQRI_FFT_LIMIT                 1651
#define __SQRI_KARATSUBA_LIMIT           36
#define __SQRR_SQRI_LIMIT                3456
#define __ZM2_MUL_LIMIT                  13
#else
#define __AGM_ATAN_LIMIT                 5088
#define __DIVRR_GMP_LIMIT                -1
#define __EXPNEWTON_LIMIT                2112
#define __EXTGCD_HALFGCD_LIMIT           83
#define __F2x_MUL_KARATSUBA_LIMIT        13
#define __F2x_MUL_MULII_LIMIT            6265
#define __F2xqX_BARRETT_LIMIT            58
#define __F2xqX_DIVREM_BARRETT_LIMIT     127
#define __F2xqX_EXTGCD_LIMIT             10
#define __F2xqX_GCD_LIMIT                257
#define __F2xqX_HALFGCD_LIMIT            43
#define __F2xqX_INVBARRETT_LIMIT         38
#define __F2xqX_REM_BARRETT_LIMIT        127
#define __Flx_BARRETT2_LIMIT             102
#define __Flx_BARRETT_LIMIT              433
#define __Flx_DIVREM2_BARRETT_LIMIT      194
#define __Flx_DIVREM_BARRETT_LIMIT       1888
#define __Flx_EXTGCD2_LIMIT              25
#define __Flx_EXTGCD_LIMIT               21
#define __Flx_GCD2_LIMIT                 2817
#define __Flx_GCD_LIMIT                  4986
#define __Flx_HALFGCD2_LIMIT             307
#define __Flx_HALFGCD_LIMIT              1245
#define __Flx_INVBARRETT2_LIMIT          1396
#define __Flx_INVBARRETT_LIMIT           1233
#define __Flx_MUL2_KARATSUBA_LIMIT       9
#define __Flx_MUL2_MULII_LIMIT           21
#define __Flx_MUL_KARATSUBA_LIMIT        57
#define __Flx_MUL_MULII_LIMIT            601
#define __Flx_REM2_BARRETT_LIMIT         296
#define __Flx_REM_BARRETT_LIMIT          1888
#define __Flx_SQR2_KARATSUBA_LIMIT       19
#define __Flx_SQR2_SQRI_LIMIT            67
#define __Flx_SQR_KARATSUBA_LIMIT        109
#define __Flx_SQR_SQRI_LIMIT             469
#define __FlxqX_BARRETT_LIMIT            17
#define __FlxqX_DIVREM_BARRETT_LIMIT     46
#define __FlxqX_EXTGCD_LIMIT             44
#define __FlxqX_GCD_LIMIT                898
#define __FlxqX_HALFGCD_LIMIT            174
#define __FlxqX_INVBARRETT_LIMIT         22
#define __FlxqX_REM_BARRETT_LIMIT        48
#define __FpXQX_BARRETT_LIMIT            12
#define __FpXQX_DIVREM_BARRETT_LIMIT     30
#define __FpXQX_EXTGCD_LIMIT             10
#define __FpXQX_GCD_LIMIT                177
#define __FpXQX_HALFGCD_LIMIT            47
#define __FpXQX_INVBARRETT_LIMIT         40
#define __FpXQX_REM_BARRETT_LIMIT        30
#define __FpX_BARRETT_LIMIT              144
#define __FpX_DIVREM_BARRETT_LIMIT       306
#define __FpX_EXTGCD_LIMIT               10
#define __FpX_GCD_LIMIT                  826
#define __FpX_HALFGCD_LIMIT              108
#define __FpX_INVBARRETT_LIMIT           337
#define __FpX_REM_BARRETT_LIMIT          306
#define __Fp_POW_BARRETT_LIMIT           97
#define __Fp_POW_REDC_LIMIT              99
#define __GCD_HALFGCD_LIMIT              13
#define __HALFGCD_LIMIT                  79
#define __INVMOD_GMP_LIMIT               -1
#define __INVNEWTON_LIMIT                2112
#define __LOGAGMCX_LIMIT                 1856
#define __LOGAGM_LIMIT                   1760
#define __MULII_FFT_LIMIT                1386
#define __MULII_KARATSUBA_LIMIT          18
#define __MULRR_MULII_LIMIT              13504
#define __RgX_MUL_LIMIT                  5
#define __RgX_SQR_LIMIT                  26
#define __SQRI_FFT_LIMIT                 1469
#define __SQRI_KARATSUBA_LIMIT           27
#define __SQRR_SQRI_LIMIT                640
#define __ZM2_MUL_LIMIT                  13
#endif
